@@ -19,12 +19,12 @@
 // DEALINGS IN THE SOFTWARE.
 
 use futures::prelude::*;
+use instant::Instant;
 use libp2p_core::{InboundUpgrade, OutboundUpgrade, UpgradeInfo};
 use libp2p_swarm::NegotiatedSubstream;
 use rand::{distributions, prelude::*};
 use std::{io, iter, time::Duration};
 use void::Void;
-use wasm_timer::Instant;
 
 /// The `Ping` protocol upgrade.
 ///
